-- Fajl sa konstantama i definicijama types i subtypes koje ce se koristiti u projektu
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

--use work.standard_additions.all;

package constants is


	-- Podesavanje trajanja ciklusa
	constant half_cycle : time := 5 ns;
	constant cycle : time := half_cycle * 2;

	constant BUS_WIDTH : natural := 32; -- velicina registara, bus-a i samog procesora
	constant DATA_WIDTH : natural := 32; -- velicina registara, word-a i samog procesora
	
	constant OPTYPE : natural := 3; -- tip instrukcije - 3 bita (pozicija 31,30,29)
	constant OPCODE : natural := 4; -- operacija koja se izvrsava - 4 bita (pozicija 28,27,26,25)
	constant REGADDR : natural := 4; -- registar koji se poziva, posto imamo 16 registara => 2^4 = 16 - 4 bita (pozicija zavisi od instrukcije) 
	constant BITFLD : natural := 1; -- odredjene instrukcije imaju samo 1 bit koji odredjuje tip instukcije ili da li je signed/unsigned
	constant SHIFT : natural := 2; -- tip rotiranja/shifrovanja - 2 bita (pozicija 6,5)
	constant COND : natural := 2; -- jump_conditionition - 2 bita (pozicija 28,27)
	constant OFFSET_LEN : natural := 26; -- duzina jump_offseta - 26 bita (pozicija 25-0)
	constant IMMEDIATE : natural := 16; -- duzina immediate - 16 bita (pozicija 15-0)
	
	constant INST_MEM_SIZE_WIDTH : natural := 6; -- stepen 2^n za velicinu memorije
	constant DATA_MEM_SIZE_WIDTH : natural := 9; -- stepen 2^n za velicinu memorije
	constant INST_MEM_SIZE : natural := 2**INST_MEM_SIZE_WIDTH; -- velicina memorije za instrukcije
	constant DATA_MEM_SIZE : natural := 2**DATA_MEM_SIZE_WIDTH; -- velicina memorije za podatke
	constant REG_COUNT : natural := 16; -- broj registara
	
	-- Pozicija specijalnih registra u GPR
	constant SP_REG_POS : natural := 13;
	constant LINK_REG_POS : natural := 14;
	constant PC_REG_POS : natural := 15;
	
	-- Pozicije statusnih bitova u CSR registru
	constant CSR_N_POS : natural := 31;
	constant CSR_Z_POS : natural := 30;
	constant CSR_C_POS : natural := 29;
	constant CSR_V_POS : natural := 28;
	
	-- Osnovna jedinica sa kojm se radi u Procesoru (moze i da se pise svuda std_ulogic_vector ali je ovako lakse
	-- Koristim ULOGIC zato sto je ne resolved tj. nema vise od 1 driver-a tj. alat ce da javi gresku ako na istu zicu pise vise od 1 elementa
	subtype ulv_t is std_ulogic_vector;
	-- Definicija za osnovni tip koji se koristi u procesoru tj. 32 bitna rec
	subtype ulv32_t is ulv_t(BUS_WIDTH - 1 downto 0);

	-- Da li je shift ili rot, smer, jump_condition
	subtype u2_t is ulv_t(SHIFT - 1 downto 0);

	--Podesavanja vezana za registar koji se adresira ("0000" = r0 itd), immediate vrednost, jump_offset, genericki bit_t
	subtype reg_t is ulv_t(REGADDR - 1 downto 0);
	subtype immed_t is ulv_t(IMMEDIATE - 1 downto 0);
	subtype offs_t is ulv_t(OFFSET_LEN - 1 downto 0);
	subtype bit_t is std_ulogic;

	-- Tip i kod instrukcije
	subtype op_type_t is ulv_t(OPTYPE - 1 downto 0);
	subtype op_code_t is ulv_t(OPCODE - 1 downto 0);
	
	-- Niz memorijskih lokacija
	type inst_mem_array_t is array(0 to INST_MEM_SIZE - 1) of ulv32_t;

	-- Niz memorijskih lokacija
	type data_mem_array_t is array(0 to DATA_MEM_SIZE - 1) of ulv32_t;
	
	-- Niz registara
	type reg_array_t is array(0 to REG_COUNT - 1) of ulv32_t;
	
	-- Enum za tip instrukcija
	type type_of_instruction_t is (NUL, DP_R, DP_I, L_S, RWM, B_BL, STOP);
	
	-- Enum za ALU Funkcije
	type alu_fn_t is (ALU_FN_NOP, ALU_FN_ADD, ALU_FN_ADDC, ALU_FN_SUB_OR_CMP, ALU_FN_SUBC, ALU_FN_AND, ALU_FN_NOT, ALU_FN_MOV);
	
	-- Enum za Adder operacije
	type adder_fn_t is (ADDER_FN_ADD, ADDER_FN_ADDC, ADDER_FN_SUB_OR_CMP, ADDER_FN_SUBC);
	
	-- Enum za Logic unit
	type logic_unit_fn_t is (LU_FN_AND, LU_FN_NOT);
	
	-- Enum za nacine detektovanja nule
	type zero_detection_t is (NOR_REDUCE, EQUALS);
	
	-- Enum za nacin detektovanja nule pomocu NOR
	type nor_reduce_t is (NOR_REDUCE, USING_LOOP);
	
	-- Enum za tipove dodavaca (adder-a)
	type adder_t is (RIPPLE_NO_ZERO_PROPAGATION, RIPPLE_WITH_ZERO_PROPAGATION, CARRY_LOOK_AHEAD, RIPPLE_FUNCTION_EQUALS_ZERO, RIPPLE_FUNCTION_NOR_ZERO);

	-- Konstanta za nulu (da ne bi svuda pisali .... = (thers => '0'); ili "00000000000000000000000000000000"
	constant ZERO_ULV32_T : ulv32_t := (others => '0');
	
	-- Konstante za nula i jedan kao bitove
	constant ZERO_BIT_T : bit_t := '0';
	constant ONE_BIT_T : bit_t := '1';

	-- Tipovi instrukcija
	constant OP_TYPE_DP_R : op_type_t := "000";
	constant OP_TYPE_DP_I : op_type_t := "001";
	constant OP_TYPE_L_S : op_type_t := "010";
	constant OP_TYPE_RWM : op_type_t := "011";
	constant OP_TYPE_B_BL : op_type_t := "100";
	constant OP_TYPE_STOP : op_type_t := "101";
	constant OP_TYPE_NUL : op_type_t := "111";

	-- ALU aritmeticke operacije
	constant OP_CODE_ADD : op_code_t := "0100";
	constant OP_CODE_ADDC : op_code_t := "0101";
	constant OP_CODE_SUB : op_code_t := "0010";
	constant OP_CODE_SUBC : op_code_t := "0110";
	
	-- ALU logicke operacije
	constant OP_CODE_AND : op_code_t := "0000";
	constant OP_CODE_NOT : op_code_t := "1111";
	constant OP_CODE_CMP : op_code_t := "1010";

	--ALU prazna operacija
	constant OP_CODE_NOP : op_code_t := "1110";
	
	-- Operacije SWAP i MOV
	constant OP_CODE_MOV : op_code_t := "1101";
	constant OP_CODE_SWAP : op_code_t := "1000";

	-- Konstanta na koju se inicijalizuje Instruction register 
	constant IR_DEFALUT_VALUE : ulv32_t := (OP_TYPE_NUL & OP_CODE_NOP & "0000000000000000000000000");

	-- Da li se radi sa signed ili unsigned brojevima
	alias UNSIGNED_OP : bit_t is ZERO_BIT_T;
	alias SIGNED_OP : bit_t is ONE_BIT_T;

	-- Tipovi load/store instrukcija
	alias L_S_S : bit_t is ZERO_BIT_T;
	alias L_S_L : bit_t is ONE_BIT_T;
	
	-- Konstante za jump_condition skoka
	constant B_BL_EQ : u2_t := "00";
	constant B_BL_GT : u2_t := "01";
	constant B_BL_HI : u2_t := "10";
	constant B_BL_AL : u2_t := "11";

	-- Tipovi branch/branch and link instrukcija
	alias B_BL_B : bit_t is ZERO_BIT_T;
	alias B_BL_BL : bit_t is ONE_BIT_T;

	-- Da li se radi sa shiftovanje ili rotiranje
	alias SHIFT_OP : bit_t is ZERO_BIT_T;
	alias ROT_OP : bit_t is ONE_BIT_T;

	-- Tipovi shift/rot opcija
	constant S_R_LL : u2_t := "00";
	constant S_R_LSR : u2_t := "01";
	constant S_R_ASR : u2_t := "10";
	constant S_R_NONE : u2_t := "11";
	
	-- Oznake za Reset zicu
	alias P_RUN : bit_t is ZERO_BIT_T;
	alias P_RST : bit_t is ONE_BIT_T;
	
	-- Definicija registara
	constant R0 : reg_t := reg_t(to_unsigned(0, reg_t'length));
	constant R1 : reg_t := reg_t(to_unsigned(1, reg_t'length));
	constant R2 : reg_t := reg_t(to_unsigned(2, reg_t'length));
	constant R3 : reg_t := reg_t(to_unsigned(3, reg_t'length));
	constant R4 : reg_t := reg_t(to_unsigned(4, reg_t'length));
	constant R5 : reg_t := reg_t(to_unsigned(5, reg_t'length));
	constant R6 : reg_t := reg_t(to_unsigned(6, reg_t'length));
	constant R7 : reg_t := reg_t(to_unsigned(7, reg_t'length));
	constant R8 : reg_t := reg_t(to_unsigned(8, reg_t'length));
	constant R9 : reg_t := reg_t(to_unsigned(9, reg_t'length));
	constant R10 : reg_t := reg_t(to_unsigned(10, reg_t'length));
	constant R11 : reg_t := reg_t(to_unsigned(11, reg_t'length));
	constant R12 : reg_t := reg_t(to_unsigned(12, reg_t'length));
	constant R13 : reg_t := reg_t(to_unsigned(13, reg_t'length));
	constant R14 : reg_t := reg_t(to_unsigned(14, reg_t'length));
	constant R15 : reg_t := reg_t(to_unsigned(15, reg_t'length));
	
	-- Specijalni registri
	constant SP_REG : reg_t := reg_t(to_unsigned(SP_REG_POS, reg_t'length));
	constant LINK_REG : reg_t := reg_t(to_unsigned(LINK_REG_POS, reg_t'length));
	constant PC_REG : reg_t := reg_t(to_unsigned(PC_REG_POS, reg_t'length));

	-- max za text operanda instrukcije je duzine 4 slova
	constant max_text_name_length : natural := 4;

	-- samo kazemo da je duzina stringa max_text_name_length
	subtype integer_to_text_name is string(1 to max_text_name_length);

	-- tabela koja sadrzi imena vezana na pozicije
	type name_table is array (natural range <>) of integer_to_text_name;
	
		-- tabela tipova instrukcija
	constant op_type_instructions_name_table : name_table(0 to 7)
	:= (
			0 => "DP_R", 
			1 => "DP_I", 
			2 => "L/S ", 
			3 => "RMW ",
			4 => "B/BL", 
			5 => "STOP",
			6 => "MISC",
			7 => "NUL "
		 );

	-- tabela ALU operacija
	constant alu_instructions_name_table : name_table(0 to 15)
	:= (
			0 => "AND ", 
			1 => "U_1 ", 
			2 => "SUB ", 
			3 => "U_3 ",
			4 => "ADD ", 
			5 => "ADC ", 
			6 => "SBC ",
			7 => "U_7 ",
			8 => "SWAP", 
			9 => "U_9 ",
		  10 => "CMP ", 
		  11 => "U_11",
		  12 => "U_12", 
		  13 => "MOV ", 
		  14 => "NOP ",
		  15 => "NOT " 
		);

	-- tabela SHIFT/ROT mogucnosti
	constant shift_rot_name_table : name_table(0 to 1)
	:= ( 
			0 => "SHFT",
			1 => "ROT " 
		);

	-- tabela SHIFT/ROT smerova
	constant shift_dir_name_table : name_table(0 to 3)
	:= (
			0 => "LL  ",
			1 => "LSR ",
			2 => "ASR ",
			3 => "NONE"
		);

	-- tabela SIGNED/UNSIGNED mogucnosti
	constant signed_unsigned_name_table : name_table(0 to 1)
	:= (
			0 => "UNSI",
			1 => "SIGN"
		);

	-- tabela COMPARISON RESULT mogucnosti
	constant comparison_results_name_table : name_table(0 to 3)
	:= (
			0 => "EQ  ",
			1 => "GT  ",
			2 => "LS  ",
			3 => "ER  "
		);

	-- tabela LOAD/STORE operacija
	constant load_store_name_table : name_table(0 to 1)
	:= (
			0 => "STR ",
			1 => "LOAD"
		);
			
	-- tabela Reset/Run stanja
	constant state_name_table : name_table(0 to 1)
	:= (
			0 => "RUN ",
			1 => "RST "
		);

	-- tabela CONDITION uslova
	constant jump_condition_name_table : name_table(0 to 3)
	:= (
			0 => "EQ  ",
			1 => "GT  ",
			2 => "HI  ",
			3 => "AL  "
		);

	-- tabela BRANCH operacija
	constant branch_name_table : name_table(0 to 1)
	:= (
			0 => "B   ",
			1 => "BL  "
		);

	-- tabela Stop operacija
	constant misc_name_table : name_table(0 to 0)
	:= (
			0 => "STOP"
		);
	
	-- provera da li je slovo whitespace
	function is_whitespace(c: character) return boolean;

	-- uklanjamo whitespace-ove na kraju string-a
	function trim_string(s: string) return string;
	
	-- konverzija bit_t u string
	function bit_t_to_string (ARG : bit_t) return string;

	-- konverzija ulv_t u unsigned string
	function to_unsigned_string (ARG : ulv_t) return string;
	
	-- konverzija ulv_t u signed string
	function to_signed_string (ARG : ulv_t) return string;

	-- konverzija jednog isecka vektora u to_bit_t
	function to_bit_t (ARG : ulv_t(0 to 0)) return bit_t;

	-- konverzija bit_t u integer
	function to_integer (ARG : bit_t) return integer;
	
	-- Konverzija ulv_t u int
	function to_integer (ARG : ulv_t) return integer;
	
	--Konverzija int kao unsigned u ulv_t
	function to_ulv32_t (ARG : integer) return ulv32_t;
	
	--Konverzija int kao signed u ulv_t
	function to_nulv32_t (ARG : integer) return ulv32_t;

end package constants;

package body constants is

	-- provera da li je slovo whitespace
	function is_whitespace(c: character) return boolean is
	begin
		if (c = ' ') or (c = HT) then
			return true;
		else return false;
		end if;
	end function;

	-- uklanjamo whitespace-ove na kraju string-a
	function trim_string(s: string) return string is
		variable j: positive;
	begin
		if (s'ascending) then
			j := s'right;
			-- pronadji prvi ne whitespace i obelezi poziciju
			for i in s'reverse_range loop
				if is_whitespace(s(i)) then
					j := j - 1;
				else exit;
				end if;
			end loop;
			return s(s'left to j);
		else
			j := s'left;
			-- pronadji prvi ne whitespace i obelezi poziciju
			for i in s'reverse_range loop
				if is_whitespace(s(i)) then
					j := j + 1;
				else exit;
				end if;
			end loop;
			return s(j downto s'right);
		end if;
	end function;
	
	-- konverzija bit_t u string
	function bit_t_to_string (ARG : bit_t) return string is
	begin
		if (ARG = '1') then
			return "TRUE";
		else
			return "FALSE";
		end if;
	end function;

	-- konverzija bit_t u string
	function to_unsigned_string (ARG : ulv_t) return string is
	begin
		return to_string(to_integer(unsigned(ARG)));
	end function;
	
	-- konverzija bit_t u string
	function to_signed_string (ARG : ulv_t) return string is
	begin
		return to_string(to_integer(signed(ARG)));
	end function;
	
	-- konverzija jednog isecka vektora u to_bit_t
	function to_bit_t (ARG : ulv_t(0 to 0)) return bit_t is
	begin
		return bit_t(ARG(0));
	end function;
	
	-- konverzija bit_t u integer
	function to_integer (ARG : bit_t) return integer is
	begin
		return to_integer(unsigned'("" & ARG));
	end function;
	
	-- Konverzija ulv_t u int
	function to_integer(ARG : ulv_t) return integer is
	begin
		return to_integer(unsigned(ARG));
	end function;
	
	--Konverzija int kao unsigned u ulv_t
	function to_ulv32_t(ARG : integer) return ulv32_t is
	begin
		return ulv32_t(to_unsigned(ARG, ulv32_t'high + 1));
	end function;
	
	--Konverzija int kao signed u ulv_t
	function to_nulv32_t(ARG : integer) return ulv32_t is
	begin
		return ulv32_t(to_signed(ARG, ulv32_t'high + 1));
	end function;

end package body constants;
