library ieee;
use ieee.std_logic_1164.std_ulogic_vector;
use STD.textio.all;
use IEEE.std_logic_1164.all;

use std.textio.all;
use ieee.numeric_std.all;


package memories is
  generic ( 
						width : positive;
            depth : positive;
            type control_type; -- za kontrolni signal
            type address_type; -- memorijska adresa
            type data_type; -- podaci
-- Konvertuje kontrolni signal u boolean da bismo znali stanje signala
            pure function "??" (c : control_type)
              return boolean is <>;
-- Rising edge ??
            function rising_edge(signal c : control_type)
              return boolean is <>;
-- Konvertuje memorijsku adresu u integer da bismo indeksirali niz
            pure function to_integer (a : address_type)
              return natural is <>;
-- Konvertovanje iz int u mem.adr. zbog load i dump
            pure function to_address_type (a : natural)
              return address_type is <>;
-- Konverzija zato sto load i dump koriste isti format koji je
-- koriscen za std_ulogic_vector
            pure function to_std_ulogic_vector (d : data_type)
              return std_ulogic_vector is <>;
            pure function to_data_type (d : std_ulogic_vector)
              return data_type is <> );
-- Reprezentacija memorije kroz niz
  type RAM_type is array (0 to 2**depth - 1) of data_type;

  procedure read_RAM (signal   RAM     : in  RAM_type;
                      constant address : in  address_type;
                      signal   data    : out data_type);

 procedure write_RAM (signal   RAM     : out RAM_type;
                       constant address : in  address_type;
                       constant data    : in  data_type);

 procedure asynch_SRAM (signal RAMin      : in  RAM_type;
												signal RAMout		 : out RAM_type;
                         signal wr       : in  control_type;
                         signal address  : in  address_type;
                         signal data_in  : out data_type;
                         signal data_out : out data_type);

   procedure flow_through_SSRAM
              (signal RAMin        : in  RAM_type;
							 signal RAMout			: out RAM_type;
               signal clk, en, wr : in  control_type;
               signal address     : in  address_type;
               signal data_in     : out data_type;
               signal data_out    : out data_type);

   procedure pipelined_SSRAM
              (signal RAM         : out  RAM_type;
               signal clk, en, wr : in  control_type;
               signal address     : in  address_type;
               signal data_in     : out data_type;
               signal data_out    : out data_type);

-- Ispisuje u fajl koji je zadat promenljivom file_name
-- Ispisuje od adrese1 do adrese2 koje su zadate parametrima start_address i finish_address
 procedure dump_RAM (signal   RAM            : in  RAM_type;
                      constant file_name      : in  string;
                      constant start_address  : in  address_type
                        := to_address_type(0);
                      constant finish_address : in  address_type
                        := to_address_type(2**depth - 1));

-- Cita iz fajla file_name i upisuje u memoriju od start_address lokacije do finish_address lokacije
-- impure function load_RAM
--                    (constant file_name      : in  string;
--                     constant start_address  : in  address_type
--                       := to_address_type(0);
--                     constant finish_address : in  address_type
--                       := to_address_type(2**depth - 1))
--                    return RAM_type;

end package memories;


package body memories is

--cita podatak sa odredjene adrese.
 procedure read_RAM (signal   RAM     : in  RAM_type;
                      constant address : in  address_type;
                      signal   data    : out data_type) is
  begin
    assert to_integer(address) <= 2**depth - 1; --provera da li je adresa u RAM-u
    data <= RAM(to_integer(address));--ako jeste, onda se u data upisuje vrednost sa te adrese
  end procedure read_RAM;

-- upisuje podatak na odredjenu adresu u RAM-u
procedure write_RAM (signal   RAM     : out RAM_type;
                       constant address : in  address_type;
                       constant data    : in  data_type) is
  begin
    assert to_integer(address) <= 2**depth - 1; --provera da li je adresa u RAM-u
    RAM(to_integer(address)) <= data; -- upis na odredjenu adresu u RAM-u
  end procedure write_RAM;

--asinhroni SRAM, upisuje kada je wr signal, inace cita
-- dodat signal RAMout jer je problem pri prozivanju f-ja read i write.
-- u slucaju read ram je tipa in, a kod write je out pa se ne moze koristiti u obe f-je
procedure asynch_SRAM (signal RAMin   	   : in  RAM_type;
												signal RAMout		 : out RAM_type;
                         signal wr       : in  control_type;
                         signal address  : in  address_type;
                         signal data_in  : out data_type;
                         signal data_out : out data_type) is
  begin
    loop
      if wr then
        write_RAM(RAMout, address, data_in);
        data_out <= data_in;
      else
        read_RAM(RAMin, address, data_out);
      end if;
      wait on wr, address, data_in;
    end loop;
  end procedure asynch_SRAM;

-- protok kroz SSRAM.
-- ako stignu signali clk, en i wr tada upisuje na adresu u ramu podatak data_in,
-- nakon cega odmah ispisuje taj podatak u data_out
-- ako nisu signali clk, en i wr tada cita iz memorije sa zadate adrese i upisuje u data_out
procedure flow_through_SSRAM
              (signal RAMin        : in  RAM_type;
							 signal RAMout			: out RAM_type;
               signal clk, en, wr : in  control_type;
               signal address     : in  address_type;
               signal data_in     : out data_type;
               signal data_out    : out data_type) is
  begin
    loop
      if rising_edge(clk) then
        if en then
          if wr then
            write_RAM(RAMout, address, data_in);
            data_out <= data_in;
          else
            read_RAM(RAMin, address, data_out);
          end if;
        end if;
      end if;
      wait on clk;
    end loop;
  end procedure flow_through_SSRAM;

-- U prvm ciklusu podatak se ucitava u registar na izlazu, 
-- a tek u drugom ciklusu se upisuje u data_out, tako da 
-- podatak iz pipelined_SSRAM dolazi za jedan ciklus kasnije nego kod flow_through_SSRAM
-- promenila sam tip RAM da je out umesto in zbog write operacije
 procedure pipelined_SSRAM
              (signal RAM         : out  RAM_type;
               signal clk, en, wr : in  control_type;
               signal address     : in  address_type;
               signal data_in     : out data_type;
               signal data_out    : out data_type) is
    variable pipelined_en       : control_type;
    variable pipelined_data_out : data_type;
  begin
    loop
      if rising_edge(clk) then
        if pipelined_en then
          data_out <= pipelined_data_out;
        end if;
        pipelined_en := en;
        if en then
          if wr then
            write_RAM(RAM, address, data_in);
            pipelined_data_out := data_in;
          else
            assert to_integer(address) <= 2**depth - 1;
            pipelined_data_out := RAM(to_integer(address));
          end if;
        end if;
      end if;
      wait on clk;
    end loop;
  end procedure pipelined_SSRAM;


-- Upisuje u fajl podatke iz memorije koji se nalaze od start_address do finish_address
 procedure dump_RAM (signal   RAM            : in  RAM_type;
                      constant file_name      : in  string;
                      constant start_address  : in  address_type
                        := to_address_type(0);
                      constant finish_address : in  address_type
                        := to_address_type(2**depth - 1)) is
    
    file dump_file  : text; --fajl u koji ce biti upisan sadrzaj memorije
    variable status : file_open_status; --da li je fajl otvoren ili ne
    variable L : line; -- pomocna linija u koju cemo upisivati mem.reci pa se iz nje prepisuje u fajl 
		-- adresa od koje se prepisuje
    constant start_address_int : natural
               := to_integer(start_address);
		-- adresa do koje se prepisuje
    constant finish_address_int : natural
               := to_integer(finish_address);
		-- trenutna adresa
    variable address : natural;
    
  begin
-- provera da li je startna adresa van memorije
    if start_address_int >= 2**depth - 1 then
      report "dump_RAM: start address "
             & to_string(start_address_int) & " out of range"
        severity error;
      return;
    end if;
    if finish_address_int <= 2**depth - 1 then
      report "dump_RAM: finish address "
             & to_string(finish_address_int) & " out of range"
        severity error;
      return;
    end if;
-- otvaramo fajl dump_file, dajemo mu ime file_name, odredjujemo da se fajl otvara za upis i koji je status
    file_open(f => dump_file, external_name => file_name,
              open_kind => write_mode, status => status);
-- provera da li se fajl uspesno otvorio i ako nije izbacivanje greske
    if status /= open_ok then
      report "dump_RAM: " & to_string(status) &
             " opening file " & file_name
        severity error;
      return;
    end if;
-- Ispis pocetne adrese
    write(L, '@');
    hwrite(L, to_unsigned(to_integer(start_address), depth));
    writeline(dump_file, L);
-- ispis podataka u fajl, prvo punimo liniju L, pa iz nje prepisujemo u fajl
    for address in to_integer(start_address)
                   to to_integer(finish_address) loop
		-- problem sa to_std_ulogic_vector(RAM(address)) !!!! 
		-- Da li sam ukljucila sva potrebna zaglavlja????
      --hwrite(L, to_std_ulogic_vector(RAM(address)));
      writeline(dump_file, L);
    end loop;
-- zatvaramo fajl nakon zavrsenog upisa
    file_close(f => dump_file);
  end procedure dump_RAM;

-- Error: C:/procesor/VHD/Memory.vhd(263): (vcom-1360) Array type of "next_address_unsigned" does not have an index constraint.
-- Error: C:/procesor/VHD/Memory.vhd(289): (vcom-1136) Unknown identifier "at_char".
--	impure function load_RAM (constant file_name : in string;
--														constant start_address : in address_type := to_address_type(0);
--														constant finish_address : in address_type:= to_address_type(2**depth - 1)) return RAM_type is
--
--	file load_file : text;
--	variable status : file_open_status;
--	variable L : line := null;
--	variable ok : boolean;
	
--	constant start_address_int : natural:= to_integer(start_address);
--	constant finish_address_int : natural:= to_integer(finish_address);
--	variable next_address : natural := to_integer(start_address);
--	variable next_address_unsigned : unsigned;
--	variable data_sulv : std_ulogic_vector(0 to width);
--	variable RAM : RAM_type;

--	begin
--		if start_address_int > 2**depth - 1 then
--			report "dump_RAM: start address " & to_string(start_address_int) & " out of range";
--			return RAM;
--		end if;

--		if finish_address_int > 2**depth - 1 then
--			report "dump_RAM: finish address "& to_string(finish_address_int)& " out of range";
--			return RAM;
--		end if;

--		file_open(f => load_file, external_name => file_name, open_kind => read_mode, status => status);

--		if status /= open_ok then
--			report "load_RAM: " & to_string(status)& " opening file " & file_name;
--			return RAM;
--		end if;
-- code to read and parse memory file contents
--		loop
--			exit when not ok or endfile(load_file);
			
--			if L(L'left) = '@' then -- address follows
--			read(L, at_char);
--			hread(L, next_address_unsigned, ok);
	
--				if not ok then
--					report "load_RAM: error reading address in file "& file_name;
--					exit;
--				end if;

--			next_address := to_integer(next_address_unsigned);

--			else
--				hread(L, data_sulv, ok);
				
--				if not ok then
--					report "load_RAM: error reading data at address "& to_string(next_address) & " in file " & file_name;
--					exit;
--				end if;

--			if next_address > finish_address_int or next_address < start_address_int then
--					report "load_RAM: address " & to_string(next_address)& " out of range " & to_string(start_address_int)& " to " & to_string(finish_address_int) & LF& " in file " & file_name;
--					ok := false;
--					exit;
--				RAM(next_address) := to_data_type(data_sulv);
--				next_address := next_address + 1;
--			end if;
--		end if;
--		end loop;

--	file_close(f => load_file);
--	return RAM;
--	end function load_RAM;

end package body memories;
