-- Glavni fajl za processor
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.control_unit_defs.all;
use work.alu_defs.all;
use work.operand_manipulation_defs.all;
use work.adder_defs.all;
use work.registers_defs.all;
use work.memory_defs.all;
use work.constants.all;

entity Processor is
	generic 
	(
		DEFAULT_INST_MEM_CONTENT : inst_mem_array_t := (others => ZERO_ULV32_T); -- Osnovna vrednost memorije instrukcija
		DEFAULT_DATA_MEM_CONTENT : data_mem_array_t := (others => ZERO_ULV32_T); -- Osnovna vrednost memorije podataka
		DEFAULT_REG_CONTENT : reg_array_t := (others => ZERO_ULV32_T); -- Osnovna vrednost registara
		ADDER_TYPE : adder_t := RIPPLE_NO_ZERO_PROPAGATION; -- Tip dodavaca koji ce se koristiti
		PRINT_COMMANDS : boolean := false -- Da li da se stampa koja se instrukcija obradjuje
	);
	port 
	(	 
		in_clk : in std_ulogic; -- ulazni klok
		in_rst : in std_ulogic -- reset signal
	);
end entity;

architecture rtl of Processor is

signal REG_LINK_IN_DATA : ulv32_t; -- vrednost koja ce se upisivati u LINK registar u slucaju Branch & Link

signal PC_DATA : ulv32_t; -- vrednost koja se upisuje u Program Counter-a

signal CSR_IN_DATA : ulv32_t; -- vrednost koja se upisuje u CSR registar
signal CSR_OUT_DATA : ulv32_t; -- vrednost CSR registra

signal IM_out_instruction : ulv32_t; -- instrukcija iz memorije koja se cuva u Instruction registar
signal IR_out_instruction : ulv32_t; -- instrukcija koja se uzima iz Instruction registra

signal DEC_out_data : decoder_out_signals; -- Podaci iz dekodera tj. podeljena instrukcija

signal CON_out_data : controller_out_signals; -- Podaci iz controllera

signal OE_out_data : ulv32_t; -- Vrednost iz prosirivaca offseta (koristi se za odredjivanje vrednosti u slucaju B/BL tipa instrukcija)
signal SE_out_data : ulv32_t; -- Prosirena vrednost immediate vrednosti u DP/I
signal SR_out_data : ulv32_t; -- Shift/Rot prvi operand u slucaju DP/R instrukcije

signal ALU_in_signals : alu_in_signals; -- Sta se salje u ALU
signal ALU_out_signals : alu_out_signals; -- Rezultati iz ALU

signal REG_in_rd : reg_t; -- Izlaz iz MUX koji kaze poziciju za Rd
signal REG_MUX_out_rd_data : ulv32_t; -- MUX koji ce da odredi sta se cuva u registar rd

signal REG_in_rd_data : ulv32_t; -- Podaci iz swap jedinice koji se upisuju u Rd
signal REG_in_rm_data : ulv32_t; -- Podaci iz swap jedinice koji se upisuju u Rm
signal REG_out_data : register_file_out_signals; -- Preuzeti podaci iz Registara

signal MEM_out_data : ulv32_t; -- Preuzet podatak iz memorije

begin

	CSR_IN_DATA <= (CSR_N_POS => ALU_out_signals.csr_n, CSR_Z_POS => ALU_out_signals.csr_z, CSR_C_POS => ALU_out_signals.csr_c, CSR_V_POS => ALU_out_signals.csr_v, others => '-');
	CSR : reg generic map (ZERO_ULV32_T) port map (in_clk, in_rst, CSR_IN_DATA, CON_out_data.csr_we, CSR_OUT_DATA);

	IM : instruction_memory generic map (DEFAULT_INST_MEM_CONTENT) port map(REG_out_data.PC_REG_data, IM_out_instruction);
	IR : reg generic map (IR_DEFALUT_VALUE) port map (in_clk, in_rst, IM_out_instruction, CON_out_data.ir_we, IR_out_instruction); 
	
	DEC : decoder port map (IR_out_instruction, DEC_out_data);
	
	REG : register_file generic map (DEFAULT_REG_CONTENT) port map(in_clk, in_rst, DEC_out_data.rn, REG_in_rd, DEC_out_data.rm, DEC_out_data.rs, REG_in_rd_data, CON_out_data.rd_we, REG_in_rm_data, CON_out_data.rm_we, PC_DATA, CON_out_data.PC_we, REG_out_data);
	
	CON : controller port map (in_clk, in_rst, (DEC_out_data.op_type, DEC_out_data.op_code, DEC_out_data.load_or_store, DEC_out_data.branch_or_branch_and_link), CON_out_data);
	
	SE : sign_extender port map(DEC_out_data.immediate, DEC_out_data.s_bit, SE_out_data);
	
	OE : offset_extender port map(DEC_out_data.jump_offset, OE_out_data);
	
	SR : shift_rot port map(DEC_out_data.s_bit, DEC_out_data.shift_or_rot, REG_out_data.rn_data, DEC_out_data.shift_rot_dir, REG_out_data.rs_data, SR_out_data);
	
	ALU_in_signals.alu_op <= CON_out_data.alu_op;
	ALU_INPUT_MUX : alu_in_mux port map(CON_out_data.op_type, REG_out_data.rn_data, MEM_out_data, SR_out_data, REG_out_data.rm_data, SE_out_data, ALU_in_signals.operator_1, ALU_in_signals.operator_2);
	ALU_in_signals.csr_c <= CSR_OUT_DATA(CSR_C_POS);

	ALU_U : entity work.alu(alu_adders_arch) generic map (ADDER_TYPE) port map (ALU_in_signals, ALU_out_signals);
	
	REG_REG_RD_MUX : reg_rd_mux port map (CON_out_data.op_type, DEC_out_data.rd, ALU_out_signals.result, MEM_out_data, REG_LINK_IN_DATA, REG_in_rd, REG_MUX_out_rd_data);
	
	DM : data_memory generic map (DEFAULT_DATA_MEM_CONTENT) port map(in_clk, in_rst, REG_out_data.rn_data, ALU_out_signals.result, CON_out_data.mem_we, MEM_out_data);
	
	L_AND_PC : link_and_pc_counter port map (CON_out_data.op_type, REG_out_data.PC_REG_data, OE_out_data, DEC_out_data.jump_condition, DEC_out_data.branch_or_branch_and_link, CSR_OUT_DATA(CSR_N_POS), CSR_OUT_DATA(CSR_Z_POS), CSR_OUT_DATA(CSR_C_POS), CSR_OUT_DATA(CSR_V_POS), REG_LINK_IN_DATA, PC_DATA);
	
	SWAP : swapper port map (CON_out_data.swap_instruction, REG_MUX_out_rd_data, REG_out_data.rm_data, REG_in_rd_data, REG_in_rm_data);

Command_Printer : if PRINT_COMMANDS 
generate
	Print : process (in_clk)

	use ieee.std_logic_textio.all;
	use std.textio.all;

	-- variabla u koju smestamo sta se ispisiuje na konzoli/u fajl
	variable debug_line : line;
	alias swrite is WRITE [LINE, STRING, SIDE, WIDTH];

	begin
	  if rising_edge(in_clk) then
		if (in_rst = '1') then
			swrite(debug_line, "Resetting");
		elsif (CON_out_data.ir_we = '1') then
			swrite(debug_line, "PC: ");
			hwrite(debug_line, (unsigned(REG_out_data.PC_REG_data) - 1));
			swrite(debug_line, " ");
			case DEC_out_data.op_type is -- Provera tipa
				when OP_TYPE_DP_R =>
					write(debug_line, trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type))), field => trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type)))'length, justified => right);
					write(debug_line, trim_string(alu_instructions_name_table(to_integer(DEC_out_data.op_code))), field => trim_string(alu_instructions_name_table(to_integer(DEC_out_data.op_code)))'length + 1, justified => right);
					swrite(debug_line, " $R");
					write(debug_line, to_integer(DEC_out_data.rn), field => to_unsigned_string(DEC_out_data.rn)'length, justified => right);
					swrite(debug_line, " $R");
					write(debug_line, to_integer(DEC_out_data.rd), field => to_unsigned_string(DEC_out_data.rd)'length, justified => right);
					write(debug_line, trim_string(signed_unsigned_name_table(to_integer(DEC_out_data.s_bit))), field => trim_string(signed_unsigned_name_table(to_integer(DEC_out_data.s_bit)))'length + 1, justified => right);
					swrite(debug_line, " $R");
					write(debug_line, to_integer(DEC_out_data.rm), field => to_unsigned_string(DEC_out_data.rm)'length, justified => right);
					swrite(debug_line, " $R");
					write(debug_line, to_integer(DEC_out_data.rs), field => to_unsigned_string(DEC_out_data.rs)'length, justified => right);
					write(debug_line, trim_string(shift_rot_name_table(to_integer(DEC_out_data.shift_or_rot))), field => trim_string(shift_rot_name_table(to_integer(DEC_out_data.shift_or_rot)))'length + 1, justified => right);
					write(debug_line, trim_string(shift_dir_name_table(to_integer(DEC_out_data.shift_rot_dir))), field => trim_string(shift_dir_name_table(to_integer(DEC_out_data.shift_rot_dir)))'length + 1, justified => right);
				when OP_TYPE_DP_I =>
					write(debug_line, trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type))), field => trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type)))'length, justified => right);
					write(debug_line, trim_string(alu_instructions_name_table(to_integer(DEC_out_data.op_code))), field => trim_string(alu_instructions_name_table(to_integer(DEC_out_data.op_code)))'length + 1, justified => right);
					swrite(debug_line, " $R");
					write(debug_line, to_integer(DEC_out_data.rn), field => to_unsigned_string(DEC_out_data.rn)'length, justified => right);
					swrite(debug_line, " $R");
					write(debug_line, to_integer(DEC_out_data.rd), field => to_unsigned_string(DEC_out_data.rd)'length, justified => right);
					write(debug_line, trim_string(signed_unsigned_name_table(to_integer(DEC_out_data.s_bit))), field => trim_string(signed_unsigned_name_table(to_integer(DEC_out_data.s_bit)))'length + 1, justified => right);
					swrite(debug_line, " #");
					write(debug_line, to_integer(SE_out_data), field => to_unsigned_string(SE_out_data)'length, justified => right);
				when OP_TYPE_L_S =>
					write(debug_line, trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type))), field => trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type)))'length, justified => right);
					write(debug_line, trim_string(load_store_name_table(to_integer(DEC_out_data.load_or_store))), field => trim_string(load_store_name_table(to_integer(DEC_out_data.load_or_store)))'length + 1, justified => right);
					swrite(debug_line, " $R");
					write(debug_line, to_integer(DEC_out_data.rn), field => to_unsigned_string(DEC_out_data.rn)'length, justified => right);
					swrite(debug_line, " $R");
					write(debug_line, to_integer(DEC_out_data.rd), field => to_unsigned_string(DEC_out_data.rd)'length, justified => right);
				when OP_TYPE_RWM =>
					write(debug_line, trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type))), field => trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type)))'length, justified => right);
					write(debug_line, trim_string(alu_instructions_name_table(to_integer(DEC_out_data.op_code))), field => trim_string(alu_instructions_name_table(to_integer(DEC_out_data.op_code)))'length + 1, justified => right);
					swrite(debug_line, " $R");
					write(debug_line, to_integer(DEC_out_data.rn), field => to_unsigned_string(DEC_out_data.rn)'length, justified => right);
					swrite(debug_line, " $R");
					write(debug_line, to_integer(DEC_out_data.rd), field => to_unsigned_string(DEC_out_data.rd)'length, justified => right);
					write(debug_line, trim_string(signed_unsigned_name_table(to_integer(DEC_out_data.s_bit))), field => trim_string(signed_unsigned_name_table(to_integer(DEC_out_data.s_bit)))'length + 1, justified => right);
					swrite(debug_line, " $R");
					write(debug_line, to_integer(DEC_out_data.rm), field => to_unsigned_string(DEC_out_data.rm)'length, justified => right);
				when OP_TYPE_B_BL =>
					write(debug_line, trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type))), field => trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type)))'length, justified => right);
					write(debug_line, trim_string(branch_name_table(to_integer(DEC_out_data.branch_or_branch_and_link))), field => trim_string(branch_name_table(to_integer(DEC_out_data.branch_or_branch_and_link)))'length + 1, justified => right);
					write(debug_line, trim_string(jump_condition_name_table(to_integer(DEC_out_data.jump_condition))), field => trim_string(jump_condition_name_table(to_integer(DEC_out_data.jump_condition)))'length + 1, justified => right);
					swrite(debug_line, " ");
					write(debug_line, to_integer(signed(DEC_out_data.jump_offset)), field => to_signed_string(DEC_out_data.jump_offset)'length, justified => right);
				when OP_TYPE_STOP =>
					write(debug_line, trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type))), field => trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type)))'length, justified => right);
				when OP_TYPE_NUL =>
					write(debug_line, trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type))), field => trim_string(op_type_instructions_name_table(to_integer(DEC_out_data.op_type)))'length, justified => right);
				when others =>
					swrite(debug_line, "!!!NEPOZNATI TIP INSTRUKCIJE!!!");
			end case;
		end if;
		writeline(output, debug_line);
	  end if;
	end process;
end generate;
  
end architecture;