-- arhitektura sa adderima

architecture alu_adders_arch of alu is

-- signali ka/iz logicke jedinice
signal logic_unit_in : logic_unit_in_signals;
signal logic_unit_out : logic_unit_out_signals;

-- signali ka/iz addera
signal adder_in : adder_in_signals;
signal adder_out : adder_out_signals;

-- rezultat podjedinice koji se cuva
signal sub_unit_result : alu_out_signals;

begin

	-- Alu ima 1 logicku jedinicu
	Logic_U : logic_unit port map (logic_unit_in, logic_unit_out);

	-- Alu ima 1 dodavac koji moze biti razlicitog tipa u slucaju s kojim je generisan
	Adder: 
	case ADDER_TYPE 
	generate
		when RIPPLE_NO_ZERO_PROPAGATION =>
			Adder_U : adder_ripple_carry_no_zero_propagation port map (adder_in, adder_out);
		when RIPPLE_WITH_ZERO_PROPAGATION =>
			Adder_U : adder_ripple_carry_with_zero_propagation port map (adder_in, adder_out);
		when CARRY_LOOK_AHEAD =>
			Adder_U : adder_carry_look_ahead port map (adder_in, adder_out);
		when RIPPLE_FUNCTION_EQUALS_ZERO =>
			Adder_U : entity work.adder_ripple_carry_function(equals) port map (adder_in, adder_out);
		when RIPPLE_FUNCTION_NOR_ZERO =>
			Adder_U : entity work.adder_ripple_carry_function(nor_reduce) port map (adder_in, adder_out);
	end generate;
	
	
	-- rezultat iz podjedinice se salje napolje
	out_s <= sub_unit_result;
  
	Perform_ALU_Function : process (all)
	
	------------------------------------------------------------------
	-- PROCEDURE ZA LOGICKU JEDINICU
	------------------------------------------------------------------

	procedure connect_logic_unit_in is -- konektuje ulaz sa logickom jedinicom
	begin
		logic_unit_in.operator_1 <= in_s.operator_1;
		logic_unit_in.operator_2 <= in_s.operator_2;
		logic_unit_in.operation <= LU_FN_AND;
	end procedure;

	procedure logic_unit_operation(operation : in logic_unit_fn_t) is -- operacija koja se izvrsava u logickoj jedinici
	begin
		logic_unit_in.operation <= operation;
	end procedure;

	procedure connect_logic_unit_out is -- povezivanje izlaza iz logicke jedinice sa izlazom iz alu
	begin
		sub_unit_result.result <= logic_unit_out.result;
		sub_unit_result.csr_n <= logic_unit_out.csr_n;
		sub_unit_result.csr_z <= logic_unit_out.csr_z;
		sub_unit_result.csr_c <= logic_unit_out.csr_c;
		sub_unit_result.csr_v <= logic_unit_out.csr_v;
	end procedure;
	------------------------------------------------------------------

	------------------------------------------------------------------
	-- PROCEDURE ZA ADDER JEDINICU
	------------------------------------------------------------------

	procedure connect_adder_in is -- povezuje ulaz sa adder-om
	begin
		adder_in.operator_1 <= in_s.operator_1;
		adder_in.operator_2 <= in_s.operator_2;
		adder_in.csr_c <= in_s.csr_c;
	end procedure;

	procedure adder_operation(operation : in adder_fn_t) is -- koja se operacija salje u adder
	begin

		case operation is
		
			when ADDER_FN_ADD => -- ADD
				adder_in.csr_c <= '0'; -- nema prekoracenja tako da se salje o kao carry

			when ADDER_FN_ADDC => -- ADDC
				null; -- ne radi se nista jer smo u connect_adder_in povezali parametar
				
			when ADDER_FN_SUB_OR_CMP => -- CMP = SUB
				adder_in.operator_2 <= NOT in_s.operator_2; -- posto je drugi komplement radi se not b + 1
				adder_in.csr_c <= '1';

			when ADDER_FN_SUBC => -- SUBC
				adder_in.operator_2 <= NOT in_s.operator_2; -- posto je drugi komplement radi se not b + not c
				adder_in.csr_c <= not in_s.csr_c; -- posto je u 2 komp a - b = a + not b + 1 => ako imamo prenos to postaje a + not b + not c 

		end case;

	end procedure;

	procedure connect_adder_out(operation : in adder_fn_t) is -- povezivanje izlaza iz addera sa izlazom iz ALU
	begin
		sub_unit_result.result <= adder_out.result;
		sub_unit_result.csr_n <= adder_out.csr_n;
		sub_unit_result.csr_z <= adder_out.csr_z;
		sub_unit_result.csr_c <= adder_out.csr_c;
		sub_unit_result.csr_v <= adder_out.csr_v;
		
		case operation is
		
			when ADDER_FN_ADD =>
				null;

			when ADDER_FN_ADDC =>
				null;
				
			when ADDER_FN_SUB_OR_CMP =>
				sub_unit_result.csr_c <= NOT adder_out.csr_c; -- ako je oduzimanje mora se raditi not c

			when ADDER_FN_SUBC =>
				sub_unit_result.csr_c <= NOT adder_out.csr_c; -- ako je oduzimanje mora se raditi not c

		end case;
		
	end procedure;
	------------------------------------------------------------------
	
	begin
		
		-- povezujemo podjedinice sa ulazima
		connect_adder_in;
		connect_logic_unit_in;
		
		case in_s.alu_op is -- Proveravamo koja je operacija

			when ALU_FN_AND => -- AND
				logic_unit_operation(LU_FN_AND);
				connect_logic_unit_out;

			when ALU_FN_NOT => -- NOT
				logic_unit_operation(LU_FN_NOT);
				connect_logic_unit_out;

			when ALU_FN_ADD => -- ADD
				adder_operation(ADDER_FN_ADD);
				connect_adder_out(ADDER_FN_ADD);

			when ALU_FN_ADDC => -- ADDC
				adder_operation(ADDER_FN_ADDC);
				connect_adder_out(ADDER_FN_ADDC);

			when ALU_FN_SUB_OR_CMP => -- SUB or CMP
				adder_operation(ADDER_FN_SUB_OR_CMP);
				connect_adder_out(ADDER_FN_SUB_OR_CMP);

			when ALU_FN_SUBC => -- SUBC
				adder_operation(ADDER_FN_SUBC);
				connect_adder_out(ADDER_FN_SUBC);
				
			when ALU_FN_MOV =>
				sub_unit_result.result <= in_s.operator_2;
				sub_unit_result.csr_n <= '0';
				sub_unit_result.csr_z <= '0';
				sub_unit_result.csr_c <= '0';
				sub_unit_result.csr_v <= '0';

			when ALU_FN_NOP =>
				sub_unit_result.result <= ZERO_ULV32_T;
				sub_unit_result.csr_n <= '0';
				sub_unit_result.csr_z <= '0';
				sub_unit_result.csr_c <= '0';
				sub_unit_result.csr_v <= '0';

		end case;
	end process;

end architecture;
