-- Fajl za kontroler koji ce na osnovu primljenih podataka iz Instruction dekodera odrediti tok procesora
library ieee;
use ieee.std_logic_1164.all;

use work.control_unit_defs.all;
use work.constants.all;

entity controller is
	port 
	(
		in_clk : in std_ulogic; -- ulazni clock
		in_rst : in std_ulogic; -- reset
		in_s : in controller_in_signals; -- ulazni signali
		out_s : out controller_out_signals -- izlazni signali
	);
end entity;

architecture rtl of controller is

	type state_type is 
						(
							S_RESET, -- resetujemo procesor
							S_IF, -- instruction fetch
							S_ID, -- instruction decode
							S_DP_X_EX, -- DP/R, DP/I izvrsavanje
							S_L_S, -- Load/Store
							--RMW, 
								S_RWM_RW, S_RMW_EX, S_RMW_MWB, -- procesi potrebni za RWM tip instrukcije
							S_B_BL, -- B/BL
							S_STOP -- STOP
						);

	signal state : state_type; -- trenutno stanje procesora

begin
	
	with in_s.op_type select -- Samo prosledjujemo enum za tip instrukcije
		out_s.op_type <=	NUL	when OP_TYPE_NUL,
							DP_R when OP_TYPE_DP_R,
							DP_I when OP_TYPE_DP_I,
							L_S	when OP_TYPE_L_S,
							RWM	when OP_TYPE_RWM,
							B_BL when OP_TYPE_B_BL,
							STOP when OP_TYPE_STOP,
							NUL	when others;

	with in_s.op_code select -- Samo prosledjujemo enum za code instrukcije
		out_s.alu_op <=	ALU_FN_NOP when OP_CODE_NOP,
						ALU_FN_ADD when OP_CODE_ADD,
						ALU_FN_ADDC when OP_CODE_ADDC,
						ALU_FN_SUB_OR_CMP when OP_CODE_SUB,
						ALU_FN_SUB_OR_CMP when OP_CODE_CMP,
						ALU_FN_SUBC when OP_CODE_SUBC,
						ALU_FN_AND when OP_CODE_AND,
						ALU_FN_NOT when OP_CODE_NOT,
						ALU_FN_NOP when others;
	
	
	process (in_clk, in_rst) -- na svaki rising edge ili reset podesavamo state machine
	begin
		if in_rst = '1' then
			state <= S_RESET; -- Reset
		elsif (rising_edge(in_clk)) then
			case state is
				-------------------------------------
				-- RESET => IF
				-------------------------------------
				when S_RESET =>
					state <= S_IF;
				-------------------------------------
				-- IF => ID
				-------------------------------------
				when S_IF =>
					state <= S_ID;
				-------------------------------------
				-- ID => odredjujemo gde dalje treba
				-------------------------------------
				when S_ID =>
					case in_s.op_type is
						----------------------
						-- DP/R
						----------------------
						when OP_TYPE_DP_R =>
							state <= S_DP_X_EX;
						-----------------------
						-- DP/I
						----------------------
						when OP_TYPE_DP_I =>
							state <= S_DP_X_EX;
						-----------------------
						-- Load/Store
						----------------------
						when OP_TYPE_L_S =>
							state <= S_L_S;
						-----------------------
						-- RWM
						----------------------
						when OP_TYPE_RWM =>
							state <= S_RWM_RW;
						-----------------------
						-- B/BL
						----------------------
						when OP_TYPE_B_BL =>
							state <= S_B_BL;
						-----------------------
						-- STOP
						----------------------
						when OP_TYPE_STOP =>
							state <= S_STOP;
						-----------------------
						-- U slucaju greske idemo u IF
						----------------------
						when others =>
							state <= S_IF;
					end case;
				-------------------------------------
				-- (DP/R or DP/I) => IF
				-------------------------------------
				when S_DP_X_EX =>
					state <= S_IF;
				-------------------------------------
				-- Load/Store => IF
				-------------------------------------
				when S_L_S =>
					state <= S_IF;
				-------------------------------------
				--------------------------------------------------------------------------------------------
				-- RWM => Upisi u Rd i idi u S_RMW_EX
				-------------------------------------
				when S_RWM_RW =>
					state <= S_RMW_EX;
				-------------------------------------
				-- S_RMW_EX => upisi u memoriju S_RMW_MWB
				-------------------------------------
				when S_RMW_EX =>
					state <= S_RMW_MWB;
				-------------------------------------
				-- S_RMW_MWB => IF
				-------------------------------------
				when S_RMW_MWB =>
					state <= S_IF;
				--------------------------------------------------------------------------------------------
				-------------------------------------
				-- B/BL => IF
				-------------------------------------
				when S_B_BL =>
					state <= S_IF;
				-------------------------------------
				-- STOP => STOP
				-------------------------------------
				when S_STOP =>
					state <= S_STOP;
			end case;
		end if;
	end process;

	process (state, in_s)
	begin
		-- samo postavljamo sve na nulu da se ne pravi latch
		out_s.rd_we <= '0';
		out_s.rm_we <= '0';
		out_s.mem_we <= '0';
		out_s.swap_instruction <= '0';
		out_s.csr_we <= '0';
		out_s.PC_we <= '0';
		out_s.ir_we <= '0';
		case state is
		
			when S_RESET =>
				null; -- reset te tim ne radimo nisa

			when S_IF =>
				out_s.ir_we <= '1'; -- upis dozvoljen u Instrukcioni registar
				
			when S_ID =>
				out_s.PC_we <= '1'; -- upis dozvoljen u PC

			when S_DP_X_EX =>
				case in_s.op_code is
					when OP_CODE_MOV => -- Ako je MOV
						out_s.rd_we <= '1'; -- upis dozvoljen u rd
					when OP_CODE_NOP =>
						null; -- NOP te ne radimo nista
					when OP_CODE_CMP => -- CMP samo podesava CSR registar
						out_s.csr_we <= '1'; -- upis dozvoljen u CSR
					when OP_CODE_SWAP => -- SWAP
						out_s.rd_we <= '1'; -- upis dozvoljen u rd
						out_s.rm_we <= '1'; -- upis dozvoljen u rm
						out_s.swap_instruction <= '1'; -- javljamo da je swap
					when others =>  -- SVE OSTALE ALU OPERACIJE
						out_s.rd_we <= '1';  -- upis dozvoljen u rd
						out_s.csr_we <= '1'; -- upis dozvoljen u CSR
				end case;

			when S_L_S =>
				out_s.rd_we <= in_s.load_or_store; -- na osnovu load_or_store bita postavljamo vrednost
				out_s.mem_we <= not in_s.load_or_store; -- na osnovu not load_or_store bita postavljamo vrednost

			when S_RWM_RW =>
				out_s.rd_we <= '1'; -- prva faza U RMW smesta vrednost iz MEM[Rn] u Rd
			when S_RMW_EX =>
				case in_s.op_code is
				
					when OP_CODE_MOV =>
						null; -- nista
					when OP_CODE_NOP =>
						null; -- nista
					when OP_CODE_CMP =>
						out_s.csr_we <= '1'; -- upis dozvoljen u CSR
					when OP_CODE_SWAP =>
						null; -- nista
					when others =>
						out_s.csr_we <= '1'; -- upis dozvoljen u CSR

				end case;
			when S_RMW_MWB =>
				out_s.mem_we <= '1'; -- tek se sad upisuje u MEM

			when S_B_BL =>
				if (in_s.branch_or_branch_and_link = B_BL_BL) then
					out_s.rd_we <= '1'; -- samo se upisuje u Rd ako je BL (vrednost koja se upisuje je LINK(
				end if;
			
			when S_STOP =>
				null; -- nista
		end case;
	end process;

end architecture;
